"""Unit tests for utils"""

from unittest import TestCase
from unittest.mock import Mock
import pytest
from elastic_transport import ApiResponseMeta
from elasticsearch9 import Elasticsearch
from elasticsearch9.exceptions import AuthenticationException, NotFoundError
from curator.exceptions import (
    ConfigurationError,
    FailedExecution,
    MissingArgument,
    RepositoryException,
    SearchableSnapshotException,
)
from curator.helpers.testers import (
    has_lifecycle_name,
    is_idx_partial,
    repository_exists,
    rollable_alias,
    snapshot_running,
    validate_filters,
    verify_client_object,
    verify_repository,
)

FAKE_FAIL = Exception('Simulated Failure')


class TestRepositoryExists(TestCase):
    """TestRepositoryExists

    Test helpers.testers.repository_exists functionality.
    """

    def test_missing_arg(self):
        """test_missing_arg

        Should raise an exception if the repository isn't passed as an arg
        """
        client = Mock()
        with pytest.raises(
            MissingArgument, match=r'No value for "repository" provided'
        ):
            repository_exists(client)

    def test_repository_in_results(self):
        """test_repository_in_results

        Should return ``True`` if the passed repository exists
        """
        client = Mock()
        client.snapshot.get_repository.return_value = {'repo': {'foo': 'bar'}}
        assert repository_exists(client, repository="repo")

    def test_repo_not_in_results(self):
        """test_repo_not_in_results

        Should return ``False`` if the passed repository does not exist
        """
        client = Mock()
        client.snapshot.get_repository.return_value = {'not_your_repo': {'foo': 'bar'}}
        assert not repository_exists(client, repository="repo")


class TestRollableAlias(TestCase):
    """TestRollableAlias

    Test helpers.testers.rollable_alias functionality.
    """

    def test_return_false_if_no_alias(self):
        """test_return_false_if_no_alias

        Should return ``False`` with a simulated Exception being raised
        """
        err = 'simulated error'
        client = Mock()
        client.info.return_value = {'version': {'number': '8.3.3'}}
        client.indices.get_alias.return_value = {}
        client.indices.get_alias.side_effect = NotFoundError(404, err, err)
        assert not rollable_alias(client, 'foo')

    def test_return_false_too_many_indices(self):
        """test_return_false_too_many_indices

        Should return ``False`` if alias associated with too many indices
        """
        retval = {
            'index-a': {'aliases': {'foo': {}}},
            'index-b': {'aliases': {'foo': {}}},
        }
        client = Mock()
        client.info.return_value = {'version': {'number': '8.3.3'}}
        client.indices.get_alias.return_value = retval
        assert not rollable_alias(client, 'foo')

    def test_return_false_non_numeric(self):
        """test_return_false_non_numeric

        Should return ``False`` if index name doesn't end in rollable digits
        """
        retval = {'index-a': {'aliases': {'foo': {}}}}
        client = Mock()
        client.info.return_value = {'version': {'number': '8.3.3'}}
        client.indices.get_alias.return_value = retval
        assert not rollable_alias(client, 'foo')

    def test_return_true_two_digits(self):
        """test_return_true_two_digits

        Should return ``True`` if the index ends in rollable digits
        """
        retval = {'index-00001': {'aliases': {'foo': {}}}}
        client = Mock()
        client.info.return_value = {'version': {'number': '8.3.3'}}
        client.indices.get_alias.return_value = retval
        assert rollable_alias(client, 'foo')

    def test_return_true_hyphenated(self):
        """test_return_true_hyphenated

        Should return ``True`` if the index has a rollable, hyphenated name
        """
        retval = {'index-2017.03.07-1': {'aliases': {'foo': {}}}}
        client = Mock()
        client.info.return_value = {'version': {'number': '8.3.3'}}
        client.indices.get_alias.return_value = retval
        assert rollable_alias(client, 'foo')


class TestSnapshotRunning(TestCase):
    """TestSnapshotRunning

    Test helpers.testers.snapshot_running functionality
    """

    # :pylint disable=line-too-long
    def test_true(self):
        """test_true

        Should return ``True`` when a snapshot is in progress/running.
        """
        client = Mock()
        client.snapshot.status.return_value = {'snapshots': ['running']}
        # self.assertTrue(snapshot_running(client))
        assert snapshot_running(client)

    def test_false(self):
        """test_False

        Should return ``False`` when a snapshot is not in progress/running.
        """
        client = Mock()
        client.snapshot.status.return_value = {'snapshots': []}
        # self.assertFalse(snapshot_running(client))
        assert not snapshot_running(client)

    def test_raises_exception(self):
        """test_raises_exception

        Should raise a ``FailedExecution`` exception when an exception happens upstream
        """
        client = Mock()
        client.snapshot.status.return_value = {'snapshots': []}
        client.snapshot.status.side_effect = FAKE_FAIL
        # self.assertRaises(FailedExecution, snapshot_running, client)
        with pytest.raises(FailedExecution, match=r'Rerun with loglevel DEBUG'):
            snapshot_running(client)


class TestValidateFilters(TestCase):
    """TestValidateFilters

    Test helpers.testers.validate_filters functionality.
    """

    def test_snapshot_with_index_filter(self):
        """test_snapshot_with_index_filter

        Should raise ConfigurationError with improper filter for filtertype
        In this case, an index filtertype (``kibana``) for the ``delete_snapshots`` filter
        """
        with pytest.raises(
            ConfigurationError, match=r'filtertype is not compatible with action'
        ):
            validate_filters('delete_snapshots', [{'filtertype': 'kibana'}])

    def test_index_with_snapshot_filter(self):
        """test_index_with_snapshot_filter

        Should raise ConfigurationError with improper filter for filtertype
        In this case, a snapshot filtertype (``state``) for the ``delete_indices`` filter
        """
        with pytest.raises(
            ConfigurationError, match=r'filtertype is not compatible with action'
        ):
            validate_filters(
                'delete_indices', [{'filtertype': 'state', 'state': 'SUCCESS'}]
            )


class TestVerifyClientObject(TestCase):
    """TestVerifyClientObject

    Test helpers.testers.verify_client_object functionality.
    """

    def test_is_client_object(self):
        """test_is_client_object

        Should return a ``None`` value for a valid client object.
        """
        test = Elasticsearch(hosts=["http://127.0.0.1:9200"])
        assert None is verify_client_object(test)

    def test_is_not_client_object(self):
        """test_is_not_client_object

        Should raise a ``TypeError`` exception with an invalid client object.
        """
        test = 'not a client object'
        with pytest.raises(TypeError, match=r'Not a valid client object'):
            verify_client_object(test)


class TestVerifyRepository(TestCase):
    """TestVerifyRepository

    Test helpers.testers.verify_repository functionality
    """

    VERIFIED_NODES = {
        'nodes': {'nodeid1': {'name': 'node1'}, 'nodeid2': {'name': 'node2'}}
    }
    REPO_NAME = 'repo_name'

    def test_passing(self):
        """test_passing

        Should return ``None`` and raise no Exception on success
        """
        client = Mock()
        client.snapshot.verify_repository.return_value = self.VERIFIED_NODES
        assert None is verify_repository(client, repository=self.REPO_NAME)

    def test_raises_404(self):
        """test_raises_404

        Should raise ``RepositoryException`` when a 404 ``TransportError`` raises first
        """
        client = Mock()
        client.snapshot.verify_repository.return_value = self.VERIFIED_NODES
        # 5 positional args for meta: status, http_version, headers, duration, node
        meta = ApiResponseMeta(404, '1.1', {}, 0.01, None)
        body = f'[{self.REPO_NAME}] missing'
        msg = 'repository_missing_exception'
        # 3 positional args for NotFoundError: message, meta, body
        effect = NotFoundError(msg, meta, body)
        client.snapshot.verify_repository.side_effect = effect
        with pytest.raises(
            RepositoryException, match=r'Repository "repo_name" not found'
        ):
            verify_repository(client, repository=self.REPO_NAME)

    def test_raises_401(self):
        """test_raises_401

        Should raise ``RepositoryException`` when a 401 AuthenticationException raises first
        """
        client = Mock()
        client.snapshot.verify_repository.return_value = self.VERIFIED_NODES
        # 5 positional args for meta: status, http_version, headers, duration, node
        meta = ApiResponseMeta(401, '1.1', {}, 0.01, None)
        body = 'No authentication'
        msg = 'authentication error'
        # 3 positional args for NotFoundError: message, meta, body
        effect = AuthenticationException(msg, meta, body)
        client.snapshot.verify_repository.side_effect = effect
        with pytest.raises(
            RepositoryException, match=r'Got a 401 response from Elasticsearch'
        ):
            verify_repository(client, repository=self.REPO_NAME)

    def test_raises_other(self):
        """test_raises_other

        Should raise ``RepositoryException`` when any other Exception raises first
        """
        client = Mock()
        client.snapshot.verify_repository.return_value = self.VERIFIED_NODES
        client.snapshot.verify_repository.side_effect = FAKE_FAIL
        with pytest.raises(RepositoryException, match=r'Failed to verify'):
            verify_repository(client, repository=self.REPO_NAME)


class TestHasLifecycleName(TestCase):
    """TestHasLifecycleName

    Test helpers.testers.has_lifecycle_name functionality
    """

    def test_has_lifecycle_name(self):
        """test_has_lifecycle_name"""
        testval = {'lifecycle': {'name': 'ilm_policy'}}
        assert has_lifecycle_name(testval)

    def test_has_no_lifecycle_name(self):
        """test_has_no_lifecycle_name"""
        testval = {'lifecycle': {'nothere': 'nope'}}
        assert not has_lifecycle_name(testval)


class TestIsIdxPartial(TestCase):
    """TestIsIdxPartial

    Test helpers.testers.is_idx_partial functionality
    """

    def test_is_idx_partial(self):
        """test_is_idx_partial"""
        testval = {'store': {'snapshot': {'partial': True}}}
        assert is_idx_partial(testval)

    def test_is_idx_partial_false1(self):
        """test_is_idx_partial_false1"""
        testval = {'store': {'snapshot': {'partial': False}}}
        assert not is_idx_partial(testval)

    def test_is_idx_partial_false2(self):
        """test_is_idx_partial_false2"""
        testval = {'store': {'snapshot': {'nothere': 'nope'}}}
        assert not is_idx_partial(testval)

    def test_is_idx_partial_raises1(self):
        """test_is_idx_partial_raises1"""
        testval = {'store': {'nothere': 'nope'}}
        with pytest.raises(
            SearchableSnapshotException, match='not a mounted searchable snapshot'
        ):
            is_idx_partial(testval)

    def test_is_idx_partial_raises2(self):
        """test_is_idx_partial_raises2"""
        testval = {'nothere': 'nope'}
        with pytest.raises(
            SearchableSnapshotException, match='not a mounted searchable snapshot'
        ):
            is_idx_partial(testval)



"""Test the Delete Indices action"""

# pylint: disable=missing-function-docstring, missing-class-docstring, line-too-long
import os
import time
import requests
from curator.helpers.getters import get_indices
from . import CuratorTestCase
from . import testvars

HOST = os.environ.get('TEST_ES_SERVER', 'http://127.0.0.1:9200')
# '      - filtertype: {0}\n'
# '        source: {1}\n'
# '        direction: {2}\n'
# '        timestring: {3}\n'
# '        unit: {4}\n'
# '        unit_count: {5}\n'
# '        field: {6}\n'
# '        stats_result: {7}\n'
# '        epoch: {8}\n')

ILM_KEYS = ['ilm-history-1-000001', 'ilm-history-1']


def exclude_ilm_history(index_list):
    """Remove any values from ILM_KEYS found in index_list"""
    for val in ILM_KEYS:
        if val in index_list:
            index_list.remove(val)
    return index_list


class TestActionFileDeleteIndices(CuratorTestCase):
    def test_retention_from_name_days(self):
        # Test extraction of unit_count from index name
        # Create indices for 10 days with retention time of 5 days in index name
        # Expected: 5 oldest indices are deleted, 5 remain
        self.args['prefix'] = 'logstash_5_'
        self.create_indices(10)
        self.write_config(self.args['configfile'], testvars.client_config.format(HOST))
        self.write_config(
            self.args['actionfile'],
            testvars.delete_pattern_proto.format(
                'age', 'name', 'older', '\'%Y.%m.%d\'', 'days', 30, '_([0-9]+)_'
            ),
        )
        self.invoke_runner()
        self.assertEqual(5, len(exclude_ilm_history(get_indices(self.client))))

    def test_retention_from_name_days_ignore_failed_match(self):
        # Test extraction of unit_count from index name
        # Create indices for 10 days with retention time of 5 days in index name
        # Create indices for 10 days with no retention time in index name
        # Expected: 5 oldest indices are deleted, 5 remain - 10 indices without
        # retention time are ignored and remain
        self.args['prefix'] = 'logstash_5_'
        self.create_indices(10)
        self.args['prefix'] = 'logstash_'
        self.create_indices(10)
        self.write_config(self.args['configfile'], testvars.client_config.format(HOST))
        self.write_config(
            self.args['actionfile'],
            testvars.delete_pattern_proto.format(
                'age',
                'name',
                'older',
                '\'%Y.%m.%d\'',
                'days',
                30,
                '_([0-9]+)_',
                ' ',
                ' ',
                ' ',
            ),
        )
        self.invoke_runner()
        self.assertEqual(15, len(exclude_ilm_history(get_indices(self.client))))

    def test_retention_from_name_days_keep_exclude_false_after_failed_match(self):
        # Test extraction of unit_count from index name and confirm correct
        # behavior after a failed regex match with no fallback time - see gh issue 1206
        # Create indices for 30 days with retention time of 5 days in index name
        #
        # Create indices for 10 days with no retention time in index name
        # that alphabetically sort before the 10 with retention time
        #
        # Create indices for 10 days with no retention time in index name
        # that sort after the 10 with retention time

        # Expected: 45 oldest matching indices are deleted, 5 matching indices remain
        # 20 indices without retention time are ignored and remain
        # overall 25 indices should remain
        self.args['prefix'] = 'logstash-aanomatch-'
        self.create_indices(10)
        self.args['prefix'] = 'logstash-match-5-'
        self.create_indices(30)
        self.args['prefix'] = 'logstash-zznomatch-'
        self.create_indices(10)
        self.write_config(self.args['configfile'], testvars.client_config.format(HOST))
        self.write_config(
            self.args['actionfile'],
            testvars.delete_pattern_proto.format(
                'age',
                'name',
                'older',
                '\'%Y.%m.%d\'',
                'days',
                -1,
                r'logstash-\w+-([0-9]+)-[0-9]{4}\.[0-9]{2}\.[0-9]{2}',
                ' ',
                ' ',
                ' ',
            ),
        )
        self.invoke_runner()
        self.assertEqual(25, len(exclude_ilm_history(get_indices(self.client))))

    def test_retention_from_name_days_failed_match_with_fallback(self):
        # Test extraction of unit_count from index name
        # Create indices for 10 days with retention time of 5 days in index name
        # Create indices for 10 days with no retention time in index name but
        # configure fallback value of 7
        # Expected: 5 oldest indices are deleted, 5 remain - 7 indices without
        # retention time are ignored and remain due to the fallback value
        self.args['prefix'] = 'logstash_5_'
        self.create_indices(10)
        self.args['prefix'] = 'logstash_'
        self.create_indices(10)
        self.write_config(self.args['configfile'], testvars.client_config.format(HOST))
        self.write_config(
            self.args['actionfile'],
            testvars.delete_pattern_proto.format(
                'age',
                'name',
                'older',
                '\'%Y.%m.%d\'',
                'days',
                7,
                '_([0-9]+)_',
                ' ',
                ' ',
                ' ',
            ),
        )
        self.invoke_runner()
        self.assertEqual(12, len(exclude_ilm_history(get_indices(self.client))))

    def test_retention_from_name_no_capture_group(self):
        # Test extraction of unit_count from index name when pattern contains no
        # capture group
        # Create indices for 10 months with retention time of 2 months in index name
        # Expected: all indices remain as the pattern cannot be used to extract a
        # retention time
        self.args['prefix'] = 'logstash_2_'
        self.args['time_unit'] = 'months'
        self.create_indices(10)
        self.write_config(self.args['configfile'], testvars.client_config.format(HOST))
        self.write_config(
            self.args['actionfile'],
            testvars.delete_pattern_proto.format(
                'age',
                'name',
                'older',
                '\'%Y.%m\'',
                'months',
                -1,
                '_[0-9]+_',
                ' ',
                ' ',
                ' ',
            ),
        )
        self.invoke_runner()
        self.assertEqual(10, len(exclude_ilm_history(get_indices(self.client))))

    def test_retention_from_name_illegal_regex_no_fallback(self):
        # Test extraction of unit_count from index name when pattern contains an
        # illegal regular expression
        # Create indices for 10 months with retention time of 2 months in index name
        # Expected: all indices remain as the pattern cannot be used to extract a
        # retention time
        self.args['prefix'] = 'logstash_2_'
        self.args['time_unit'] = 'months'
        self.create_indices(10)
        self.write_config(self.args['configfile'], testvars.client_config.format(HOST))
        self.write_config(
            self.args['actionfile'],
            testvars.delete_pattern_proto.format(
                'age',
                'name',
                'older',
                '\'%Y.%m\'',
                'months',
                -1,
                '_[0-9+_',
                ' ',
                ' ',
                ' ',
            ),
        )
        self.invoke_runner()
        self.assertEqual(10, len(exclude_ilm_history(get_indices(self.client))))

    def test_retention_from_name_illegal_regex_with_fallback(self):
        # Test extraction of unit_count from index name when pattern contains an
        # illegal regular expression
        # Create indices for 10 days with retention time of 2 days in index name
        # Expected: Fallback value of 3 is used and 3 most recent indices remain
        # in place
        self.args['prefix'] = 'logstash_2_'
        self.create_indices(10)
        self.write_config(self.args['configfile'], testvars.client_config.format(HOST))
        self.write_config(
            self.args['actionfile'],
            testvars.delete_pattern_proto.format(
                'age',
                'name',
                'older',
                '\'%Y.%m.%d\'',
                'days',
                3,
                '_[0-9+_',
                ' ',
                ' ',
                ' ',
            ),
        )
        self.invoke_runner()
        self.assertEqual(3, len(exclude_ilm_history(get_indices(self.client))))

    def test_name_older_than_now(self):
        self.create_indices(10)
        self.write_config(self.args['configfile'], testvars.client_config.format(HOST))
        self.write_config(
            self.args['actionfile'],
            testvars.delete_proto.format(
                'age', 'name', 'older', '\'%Y.%m.%d\'', 'days', 5, ' ', ' ', ' '
            ),
        )
        self.invoke_runner()
        self.assertEqual(5, len(exclude_ilm_history(get_indices(self.client))))

    def test_name_based_age_match_five(self):
        self.create_indices(10)
        self.create_index(name='foobert')
        self.create_index(name='.shouldbehidden')
        self.create_index(name='this_has_no_date')
        self.write_config(self.args['configfile'], testvars.client_config.format(HOST))
        self.write_config(
            self.args['actionfile'],
            testvars.delete_proto.format(
                'age', 'name', 'older', '\'%Y.%m.%d\'', 'days', 5, ' ', ' ', ' '
            ),
        )
        self.invoke_runner()
        self.assertEqual(8, len(exclude_ilm_history(get_indices(self.client))))

    def test_creation_date_newer_than_epoch(self):
        self.create_indices(10)
        self.write_config(self.args['configfile'], testvars.client_config.format(HOST))
        self.write_config(
            self.args['actionfile'],
            testvars.delete_proto.format(
                'age',
                'creation_date',
                'younger',
                ' ',
                'seconds',
                0,
                ' ',
                ' ',
                int(time.time()) - 60,
            ),
        )
        self.invoke_runner()
        self.assertEqual(0, len(exclude_ilm_history(get_indices(self.client))))

    def test_delete_in_period(self):
        # filtertype: {0}
        # source: {1}
        # range_from: {2}
        # range_to: {3}
        # timestring: {4}
        # unit: {5}
        # field: {6}
        # stats_result: {7}
        # intersect: {8}
        # epoch: {9}
        # week_starts_on: {10}
        self.create_indices(10)
        self.write_config(self.args['configfile'], testvars.client_config.format(HOST))
        self.write_config(
            self.args['actionfile'],
            testvars.delete_period_proto.format(
                'period',
                'name',
                '-5',
                '-1',
                "'%Y.%m.%d'",
                'days',
                ' ',
                ' ',
                ' ',
                ' ',
                'monday',
            ),
        )
        self.invoke_runner()
        self.assertEqual(0, self.result.exit_code)
        self.assertEqual(5, len(exclude_ilm_history(get_indices(self.client))))

    def test_delete_in_period_absolute_date(self):
        delete_period_abs = (
            '---\n'
            'actions:\n'
            '  1:\n'
            '    description: "Delete indices as filtered"\n'
            '    action: delete_indices\n'
            '    options:\n'
            '      continue_if_exception: False\n'
            '      disable_action: False\n'
            '    filters:\n'
            '    - filtertype: {0}\n'
            '      period_type: absolute\n'
            '      source: {1}\n'
            '      date_from: {2}\n'
            '      date_to: {3}\n'
            '      timestring: {4}\n'
            '      unit: {5}\n'
            '      date_from_format: {6}\n'
            '      date_to_format: {7}\n'
        )
        expected = 'index-2017.02.02'
        self.create_index('index-2017.01.02')
        self.create_index('index-2017.01.03')
        self.create_index(expected)
        self.write_config(self.args['configfile'], testvars.client_config.format(HOST))
        self.write_config(
            self.args['actionfile'],
            delete_period_abs.format(
                'period',
                'name',
                '2017.01.01',
                '2017.01.10',
                "'%Y.%m.%d'",
                'days',
                "'%Y.%m.%d'",
                "'%Y.%m.%d'",
            ),
        )
        self.invoke_runner()
        indices = exclude_ilm_history(get_indices(self.client))
        self.assertEqual(0, self.result.exit_code)
        self.assertEqual(1, len(indices))
        self.assertEqual(expected, indices[0])

    def test_delete_in_period_intersect(self):
        # filtertype: {0}
        # source: {1}
        # range_from: {2}
        # range_to: {3}
        # timestring: {4}
        # unit: {5}
        # field: {6}
        # stats_result: {7}
        # intersect: {8}
        # epoch: {9}
        # week_starts_on: {10}
        # 2017-09-01T01:00:00 = 1504227600
        # 2017-09-25T01:00:00 = 1506301200
        # 2017-09-29T01:00:00 = 1506646800
        idx1, idx2 = ('intersecting', 'notintersecting')
        self.create_index(idx1)
        self.create_index(idx2)
        self.client.index(
            index=idx1,
            id='1',
            document={'@timestamp': '2017-09-25T01:00:00Z', 'doc': 'Earliest'},
        )
        self.client.index(
            index=idx1,
            id='2',
            document={'@timestamp': '2017-09-29T01:00:00Z', 'doc': 'Latest'},
        )
        self.client.index(
            index=idx2,
            id='1',
            document={'@timestamp': '2017-09-01T01:00:00Z', 'doc': 'Earliest'},
        )
        self.client.index(
            index=idx2,
            id='2',
            document={'@timestamp': '2017-09-29T01:00:00Z', 'doc': 'Latest'},
        )
        # Decorators cause this pylint error
        # pylint: disable=E1123
        self.client.indices.flush(index='*', force=True)
        self.client.indices.refresh(index=f'{idx1},{idx2}')
        self.write_config(self.args['configfile'], testvars.client_config.format(HOST))
        self.write_config(
            self.args['actionfile'],
            testvars.delete_period_proto.format(
                'period',
                'field_stats',
                '0',
                '0',
                ' ',
                'weeks',
                "'@timestamp'",
                'min_value',
                'true',
                1506716040,
                'sunday',
            ),
        )
        self.invoke_runner()
        self.assertEqual(0, self.result.exit_code)
        indices = exclude_ilm_history(get_indices(self.client))
        self.assertEqual(1, len(indices))
        self.assertEqual(idx2, indices[0])

    def test_empty_list(self):
        self.create_indices(10)
        self.write_config(self.args['configfile'], testvars.client_config.format(HOST))
        self.write_config(
            self.args['actionfile'],
            testvars.delete_proto.format(
                'age',
                'creation_date',
                'older',
                ' ',
                'days',
                90,
                ' ',
                ' ',
                int(time.time()),
            ),
        )
        self.invoke_runner()
        self.assertEqual(10, len(exclude_ilm_history(get_indices(self.client))))
        self.assertEqual(1, self.result.exit_code)

    def test_ignore_empty_list(self):
        self.create_indices(10)
        self.write_config(self.args['configfile'], testvars.client_config.format(HOST))
        self.write_config(
            self.args['actionfile'],
            testvars.delete_ignore_proto.format(
                'age',
                'creation_date',
                'older',
                ' ',
                'days',
                90,
                ' ',
                ' ',
                int(time.time()),
            ),
        )
        self.invoke_runner()
        self.assertEqual(10, len(exclude_ilm_history(get_indices(self.client))))
        self.assertEqual(0, self.result.exit_code)

    def test_extra_options(self):
        self.write_config(self.args['configfile'], testvars.client_config.format(HOST))
        self.write_config(
            self.args['actionfile'],
            testvars.bad_option_proto_test.format('delete_indices'),
        )
        self.invoke_runner()
        self.assertEqual(1, self.result.exit_code)

    def test_945(self):
        self.create_indices(10)
        self.write_config(self.args['configfile'], testvars.client_config.format(HOST))
        self.write_config(self.args['actionfile'], testvars.test_945)
        self.invoke_runner()
        self.assertEqual(1, self.result.exit_code)

    def test_name_epoch_zero(self):
        self.create_index('epoch_zero-1970.01.01')
        self.write_config(self.args['configfile'], testvars.client_config.format(HOST))
        self.write_config(
            self.args['actionfile'],
            testvars.delete_proto.format(
                'age', 'name', 'older', '\'%Y.%m.%d\'', 'days', 5, ' ', ' ', ' '
            ),
        )
        self.invoke_runner()
        self.assertEqual(0, len(exclude_ilm_history(get_indices(self.client))))

    def test_name_negative_epoch(self):
        self.create_index('index-1969.12.31')
        self.write_config(self.args['configfile'], testvars.client_config.format(HOST))
        self.write_config(
            self.args['actionfile'],
            testvars.delete_proto.format(
                'age', 'name', 'older', '\'%Y.%m.%d\'', 'days', 5, ' ', ' ', ' '
            ),
        )
        self.invoke_runner()
        self.assertEqual(0, len(exclude_ilm_history(get_indices(self.client))))

    def test_allow_ilm_indices_true(self):
        name = 'test'
        policy = {
            'policy': {
                'phases': {
                    'hot': {
                        'min_age': '0ms',
                        'actions': {'rollover': {'max_age': '2h', 'max_docs': 4}},
                    }
                }
            }
        }
        url = f'{HOST}/_ilm/policy/{name}'
        _ = requests.put(url, json=policy, timeout=30)
        # print(r.text) # logging reminder
        self.create_indices(10, ilm_policy=name)
        self.write_config(self.args['configfile'], testvars.client_config.format(HOST))
        self.write_config(
            self.args['actionfile'],
            testvars.ilm_delete_proto.format(
                'age', 'name', 'older', '\'%Y.%m.%d\'', 'days', 5, ' ', ' ', ' ', 'true'
            ),
        )
        self.invoke_runner()
        self.assertEqual(5, len(exclude_ilm_history(get_indices(self.client))))

    def test_allow_ilm_indices_false(self):
        name = 'test'
        policy = {
            'policy': {
                'phases': {
                    'hot': {
                        'min_age': '0ms',
                        'actions': {'rollover': {'max_age': '2h', 'max_docs': 4}},
                    }
                }
            }
        }
        url = f'{HOST}/_ilm/policy/{name}'
        _ = requests.put(url, json=policy, timeout=30)
        # print(r.text) # logging reminder
        self.create_indices(10, ilm_policy=name)
        self.write_config(self.args['configfile'], testvars.client_config.format(HOST))
        self.write_config(
            self.args['actionfile'],
            testvars.ilm_delete_proto.format(
                'age',
                'name',
                'older',
                '\'%Y.%m.%d\'',
                'days',
                5,
                ' ',
                ' ',
                ' ',
                'false',
            ),
        )
        self.invoke_runner()
        self.assertEqual(10, len(exclude_ilm_history(get_indices(self.client))))


class TestCLIDeleteIndices(CuratorTestCase):
    def test_name_older_than_now_cli(self):
        self.create_indices(10)
        args = self.get_runner_args()
        args += [
            '--config',
            self.args['configfile'],
            'delete-indices',
            '--filter_list',
            (
                '{"filtertype":"age","source":"name","direction":"older",'
                '"timestring":"%Y.%m.%d","unit":"days","unit_count":5}'
            ),
        ]
        self.assertEqual(
            0,
            self.run_subprocess(
                args, logname='TestCLIDeleteIndices.test_name_older_than_now_cli'
            ),
        )
        self.assertEqual(5, len(exclude_ilm_history(get_indices(self.client))))



"""Test setup"""

# pylint: disable=C0115, C0116
import logging
import os
import random
import shutil
import string
import sys
import tempfile
import time
import json
import warnings
from datetime import timedelta, datetime, date, timezone
from subprocess import Popen, PIPE
from unittest import SkipTest, TestCase
from elasticsearch9 import Elasticsearch
from elasticsearch9.exceptions import ConnectionError as ESConnectionError
from elasticsearch9.exceptions import ElasticsearchWarning, NotFoundError
from click import testing as clicktest
from es_client.utils import get_version
from curator.cli import cli

from . import testvars


client = None

DATEMAP = {
    'months': '%Y.%m',
    'weeks': '%Y.%W',
    'days': '%Y.%m.%d',
    'hours': '%Y.%m.%d.%H',
}

HOST = os.environ.get('TEST_ES_SERVER', 'http://127.0.0.1:9200')


def random_directory():
    dirname = ''.join(
        random.choice(string.ascii_uppercase + string.digits) for _ in range(8)
    )
    directory = tempfile.mkdtemp(suffix=dirname)
    if not os.path.exists(directory):
        os.makedirs(directory)
    return directory


def get_client():
    # pylint: disable=global-statement, invalid-name
    global client
    if client is not None:
        return client

    client = Elasticsearch(hosts=HOST, request_timeout=300)

    # wait for yellow status
    for _ in range(100):
        time.sleep(0.1)
        try:
            # pylint: disable=E1123
            client.cluster.health(wait_for_status='yellow')
            return client
        except ESConnectionError:
            continue
    # timeout
    raise SkipTest("Elasticsearch failed to start.")


def setup():
    get_client()


class Args(dict):
    def __getattr__(self, att_name):
        return self.get(att_name, None)


class CuratorTestCase(TestCase):
    def setUp(self):
        super(CuratorTestCase, self).setUp()
        self.logger = logging.getLogger('CuratorTestCase.setUp')
        self.client = get_client()

        args = {}
        args['HOST'] = HOST
        args['time_unit'] = 'days'
        args['prefix'] = 'logstash-'
        self.args = args
        # dirname = ''.join(random.choice(string.ascii_uppercase + string.digits)
        #   for _ in range(8))
        # This will create a psuedo-random temporary directory on the machine
        # which runs the unit tests, but NOT on the machine where elasticsearch
        # is running. This means tests may fail if run against remote instances
        # unless you explicitly set `self.args['location']` to a proper spot
        # on the target machine.
        # self.args['location'] = random_directory()
        nodesinfo = self.client.nodes.info()
        nodename = list(nodesinfo['nodes'].keys())[0]
        if 'repo' in nodesinfo['nodes'][nodename]['settings']['path']:
            if isinstance(
                nodesinfo['nodes'][nodename]['settings']['path']['repo'], list
            ):
                self.args['location'] = nodesinfo['nodes'][nodename]['settings'][
                    'path'
                ]['repo'][0]
            else:
                self.args['location'] = nodesinfo['nodes'][nodename]['settings'][
                    'path'
                ]['repo']
        else:  # Use a random directory if repo is not specified, but log it
            self.logger.warning('path.repo is not configured!')
            self.args['location'] = random_directory()
        self.args['configdir'] = random_directory()
        self.args['configfile'] = os.path.join(self.args['configdir'], 'curator.yml')
        self.args['actionfile'] = os.path.join(self.args['configdir'], 'actions.yml')
        self.args['repository'] = 'test_repository'
        # if not os.path.exists(self.args['location']):
        #     os.makedirs(self.args['location'])
        self.logger.debug('setUp completed...')
        self.runner = clicktest.CliRunner()
        self.runner_args = [
            '--config',
            self.args['configfile'],
            self.args['actionfile'],
        ]
        self.result = None

    def get_version(self):
        return get_version(self.client)

    def tearDown(self):
        self.logger = logging.getLogger('CuratorTestCase.tearDown')
        self.logger.debug('tearDown initiated...')
        # re-enable shard allocation for next tests
        enable_allocation = json.loads('{"cluster.routing.allocation.enable":null}')
        self.client.cluster.put_settings(transient=enable_allocation)
        self.delete_repositories()
        # 8.0 removes our ability to purge with wildcards...
        # ElasticsearchWarning: this request accesses system indices: [.tasks],
        # but in a future major version, direct access to system indices will be
        # prevented by default
        warnings.filterwarnings("ignore", category=ElasticsearchWarning)
        indices = list(
            self.client.indices.get(index="*", expand_wildcards='open,closed').keys()
        )
        if len(indices) > 0:
            # ElasticsearchWarning: this request accesses system indices: [.tasks],
            # but in a future major version, direct access to system indices will be
            # prevented by default
            warnings.filterwarnings("ignore", category=ElasticsearchWarning)
            self.client.indices.delete(index=','.join(indices))
        for path_arg in ['location', 'configdir']:
            if os.path.exists(self.args[path_arg]):
                shutil.rmtree(self.args[path_arg])

    def parse_args(self):
        return Args(self.args)

    def create_indices(self, count, unit=None, ilm_policy=None):
        now = datetime.now(timezone.utc)
        unit = unit if unit else self.args['time_unit']
        fmt = DATEMAP[unit]
        if not unit == 'months':
            step = timedelta(**{unit: 1})
            for _ in range(count):
                self.create_index(
                    self.args['prefix'] + now.strftime(fmt),
                    wait_for_yellow=False,
                    ilm_policy=ilm_policy,
                )
                now -= step
        else:  # months
            now = date.today()
            d = date(now.year, now.month, 1)
            self.create_index(
                self.args['prefix'] + now.strftime(fmt),
                wait_for_yellow=False,
                ilm_policy=ilm_policy,
            )

            for _ in range(1, count):
                if d.month == 1:
                    d = date(d.year - 1, 12, 1)
                else:
                    d = date(d.year, d.month - 1, 1)
                self.create_index(
                    self.args['prefix'] + datetime(d.year, d.month, 1).strftime(fmt),
                    wait_for_yellow=False,
                    ilm_policy=ilm_policy,
                )
        # pylint: disable=E1123
        self.client.cluster.health(wait_for_status='yellow')

    def wfy(self):
        # pylint: disable=E1123
        self.client.cluster.health(wait_for_status='yellow')

    def create_index(
        self,
        name,
        shards=1,
        wait_for_yellow=True,
        ilm_policy=None,
        wait_for_active_shards=1,
    ):
        request_body = {'index': {'number_of_shards': shards, 'number_of_replicas': 0}}
        if ilm_policy is not None:
            request_body['index']['lifecycle'] = {'name': ilm_policy}
        # ElasticsearchWarning: index name [.shouldbehidden] starts with a dot '.',
        # in the next major version, index names starting with a dot are reserved
        # for hidden indices and system indices
        warnings.filterwarnings("ignore", category=ElasticsearchWarning)
        self.client.indices.create(
            index=name,
            settings=request_body,
            wait_for_active_shards=wait_for_active_shards,
        )
        if wait_for_yellow:
            self.wfy()

    def add_docs(self, idx):
        for i in ["1", "2", "3"]:
            self.client.create(index=idx, id=i, document={"doc" + i: 'TEST DOCUMENT'})
            # This should force each doc to be in its own segment.
            # pylint: disable=E1123
            self.client.indices.flush(index=idx, force=True)
            self.client.indices.refresh(index=idx)

    def create_snapshot(self, name, csv_indices):
        self.create_repository()
        self.client.snapshot.create(
            repository=self.args['repository'],
            snapshot=name,
            ignore_unavailable=False,
            include_global_state=True,
            partial=False,
            indices=csv_indices,
            wait_for_completion=True,
        )

    def delete_snapshot(self, name):
        try:
            self.client.snapshot.delete(
                repository=self.args['repository'], snapshot=name
            )
        except NotFoundError:
            pass

    def create_repository(self):
        request_body = {'type': 'fs', 'settings': {'location': self.args['location']}}
        self.client.snapshot.create_repository(
            name=self.args['repository'], body=request_body
        )

    def delete_repositories(self):
        result = []
        try:
            result = self.client.snapshot.get_repository(name='*')
        except NotFoundError:
            pass
        for repo in result:
            try:
                cleanup = self.client.snapshot.get(repository=repo, snapshot='*')
            # pylint: disable=broad-except
            except Exception:
                cleanup = {'snapshots': []}
            for listitem in cleanup['snapshots']:
                self.delete_snapshot(listitem['snapshot'])
            self.client.snapshot.delete_repository(name=repo)

    def close_index(self, name):
        self.client.indices.close(index=name)

    def write_config(self, fname, data):
        with open(fname, 'w', encoding='utf-8') as fhandle:
            fhandle.write(data)

    def get_runner_args(self):
        self.write_config(self.args['configfile'], testvars.client_config.format(HOST))
        runner = os.path.join(os.getcwd(), 'run_singleton.py')
        return [sys.executable, runner]

    def run_subprocess(self, args, logname='subprocess'):
        local_logger = logging.getLogger(logname)
        p = Popen(args, stderr=PIPE, stdout=PIPE)
        stdout, stderr = p.communicate()
        local_logger.debug('STDOUT = %s', stdout.decode('utf-8'))
        local_logger.debug('STDERR = %s', stderr.decode('utf-8'))
        return p.returncode

    def invoke_runner(self, dry_run=False):
        if dry_run:
            self.result = self.runner.invoke(
                cli,
                [
                    '--config',
                    self.args['configfile'],
                    '--dry-run',
                    self.args['actionfile'],
                ],
            )
            return
        self.result = self.runner.invoke(cli, self.runner_args)

    def invoke_runner_alt(self, **kwargs):
        myargs = []
        if kwargs:
            for key, value in kwargs.items():
                myargs.append(f'--{key}')
                myargs.append(value)
            myargs.append(self.args['actionfile'])
            self.result = self.runner.invoke(cli, myargs)



"""Test CLI functionality"""

# pylint: disable=C0115, C0116, invalid-name
import os
from curator.exceptions import ConfigurationError
from curator.helpers.getters import get_indices
from . import CuratorTestCase
from . import testvars

HOST = os.environ.get('TEST_ES_SERVER', 'http://127.0.0.1:9200')


class TestCLIMethods(CuratorTestCase):
    def test_bad_client_config(self):
        self.create_indices(10)
        self.write_config(
            self.args['configfile'], testvars.bad_client_config.format(HOST)
        )
        self.write_config(
            self.args['actionfile'],
            testvars.disabled_proto.format('close', 'delete_indices'),
        )
        self.invoke_runner()
        assert 1 == self.result.exit_code

    def test_cli_client_config(self):
        self.create_indices(10)
        self.write_config(
            self.args['configfile'], testvars.bad_client_config.format(HOST)
        )
        self.write_config(
            self.args['actionfile'],
            testvars.disabled_proto.format('close', 'delete_indices'),
        )
        self.invoke_runner_alt(
            hosts='http://127.0.0.1:9200', loglevel='DEBUG', logformat='ecs'
        )
        assert 0 == self.result.exit_code

    def test_cli_unreachable_cloud_id(self):
        self.create_indices(10)
        self.write_config(
            self.args['actionfile'],
            testvars.disabled_proto.format('close', 'delete_indices'),
        )
        self.invoke_runner_alt(cloud_id='abc:def', username='user', password='pass')
        assert 1 == self.result.exit_code

    def test_no_config(self):
        # This test checks whether localhost:9200 is provided if no hosts or
        # port are in the configuration. But in testing, sometimes
        # TEST_ES_SERVER is set to something other than localhost:9200.  In this
        # case, the test here would fail.  The if statement at the end now
        # compensates. See https://github.com/elastic/curator/issues/843
        localtest = False
        if HOST == 'http://127.0.0.1:9200':
            localtest = True
        self.create_indices(10)
        self.write_config(self.args['configfile'], '---\n')  # Empty YAML file.
        self.write_config(
            self.args['actionfile'],
            testvars.disabled_proto.format('close', 'delete_indices'),
        )
        self.invoke_runner()
        if localtest:
            assert 0 == self.result.exit_code
        else:
            assert -1 == self.result.exit_code

    def test_no_logging_config(self):
        self.create_indices(10)
        self.write_config(
            self.args['configfile'], testvars.no_logging_config.format(HOST)
        )
        self.write_config(
            self.args['actionfile'],
            testvars.disabled_proto.format('close', 'delete_indices'),
        )
        self.invoke_runner()
        assert 0 == self.result.exit_code

    def test_logging_none(self):
        self.create_indices(10)
        self.write_config(
            self.args['configfile'], testvars.none_logging_config.format(HOST)
        )
        self.write_config(
            self.args['actionfile'],
            testvars.disabled_proto.format('close', 'delete_indices'),
        )
        self.invoke_runner()
        assert 0 == self.result.exit_code

    def test_invalid_action(self):
        self.write_config(self.args['configfile'], testvars.client_config.format(HOST))
        self.write_config(
            self.args['actionfile'], testvars.optionless_proto.format('invalid_action')
        )
        self.invoke_runner()
        assert 1 == self.result.exit_code

    def test_action_is_none(self):
        self.write_config(self.args['configfile'], testvars.client_config.format(HOST))
        self.write_config(
            self.args['actionfile'], testvars.optionless_proto.format(' ')
        )
        self.invoke_runner()
        assert isinstance(self.result.exception, ConfigurationError)

    def test_no_action(self):
        self.write_config(self.args['configfile'], testvars.client_config.format(HOST))
        self.write_config(self.args['actionfile'], testvars.actionless_proto)
        self.invoke_runner()
        assert isinstance(self.result.exception, ConfigurationError)

    def test_dry_run(self):
        self.create_indices(10)
        self.write_config(self.args['configfile'], testvars.client_config.format(HOST))
        self.write_config(
            self.args['actionfile'],
            testvars.delete_proto.format(
                'age', 'name', 'older', '\'%Y.%m.%d\'', 'days', 5, ' ', ' ', ' '
            ),
        )
        self.invoke_runner(dry_run=True)
        assert 10 == len(get_indices(self.client))

    def test_action_disabled(self):
        self.create_indices(10)
        self.write_config(self.args['configfile'], testvars.client_config.format(HOST))
        self.write_config(
            self.args['actionfile'],
            testvars.disabled_proto.format('close', 'delete_indices'),
        )
        self.invoke_runner()
        assert 0 == len(get_indices(self.client))
        assert 0 == self.result.exit_code

    # I'll have to think up another way to create an exception.
    # The exception that using "alias" created, a missing argument,
    # is caught too early for this to actually run the test now :/
    def test_continue_if_exception(self):
        name = 'log1'
        self.create_index(name)
        self.create_index('log2')
        self.write_config(self.args['configfile'], testvars.client_config.format(HOST))
        self.write_config(
            self.args['actionfile'],
            testvars.continue_proto.format(name, True, 'delete_indices', False),
        )
        self.invoke_runner()
        assert 0 == len(get_indices(self.client))
        assert 0 == self.result.exit_code

    def test_continue_if_exception_false(self):
        name = 'log1'
        self.create_index(name)
        self.create_index('log2')
        self.write_config(self.args['configfile'], testvars.client_config.format(HOST))
        self.write_config(
            self.args['actionfile'],
            testvars.continue_proto.format(name, False, 'delete_indices', False),
        )
        self.invoke_runner()
        assert 2 == len(get_indices(self.client))
        assert 1 == self.result.exit_code

    def test_no_options_in_action(self):
        self.create_indices(10)
        self.create_index('my_index')  # Added for the ILM filter's sake
        self.write_config(self.args['configfile'], testvars.client_config.format(HOST))
        self.write_config(
            self.args['actionfile'], testvars.no_options_proto.format('delete_indices')
        )
        self.invoke_runner(dry_run=True)
        assert 0 == self.result.exit_code



from elasticsearch9 import ConflictError, NotFoundError, TransportError

fake_fail = Exception('Simulated Failure')
four_oh_one = TransportError(401, "simulated error")
four_oh_four = TransportError(404, "simulated error")
get_alias_fail = NotFoundError(404, 'simulated error', 'simulated error')
named_index = 'index_name'
named_indices = ["index-2015.01.01", "index-2015.02.01"]
open_index = {'metadata': {'indices': {named_index: {'state': 'open'}}}}
closed_index = {'metadata': {'indices': {named_index: {'state': 'close'}}}}
cat_open_index = [{'status': 'open'}]
cat_closed_index = [{'status': 'close'}]
open_indices = {
    'metadata': {'indices': {'index1': {'state': 'open'}, 'index2': {'state': 'open'}}}
}
closed_indices = {
    'metadata': {
        'indices': {'index1': {'state': 'close'}, 'index2': {'state': 'close'}}
    }
}
named_alias = 'alias_name'
alias_retval = {"pre_aliased_index": {"aliases": {named_alias: {}}}}
rollable_alias = {"index-000001": {"aliases": {named_alias: {}}}}
rollover_conditions = {'conditions': {'max_age': '1s'}}
dry_run_rollover = {
    "acknowledged": True,
    "shards_acknowledged": True,
    "old_index": "index-000001",
    "new_index": "index-000002",
    "rolled_over": False,
    "dry_run": True,
    "conditions": {"max_age": "1s"},
}
aliases_retval = {
    "index1": {"aliases": {named_alias: {}}},
    "index2": {"aliases": {named_alias: {}}},
}
alias_one_add = [{'add': {'alias': 'alias', 'index': 'index_name'}}]
alias_one_add_with_extras = [
    {
        'add': {
            'alias': 'alias',
            'index': 'index_name',
            'filter': {'term': {'user': 'kimchy'}},
        }
    }
]
alias_one_rm = [{'remove': {'alias': 'my_alias', 'index': named_index}}]
alias_one_body = {
    "actions": [
        {'remove': {'alias': 'alias', 'index': 'index_name'}},
        {'add': {'alias': 'alias', 'index': 'index_name'}},
    ]
}
alias_two_add = [
    {'add': {'alias': 'alias', 'index': 'index-2016.03.03'}},
    {'add': {'alias': 'alias', 'index': 'index-2016.03.04'}},
]
alias_two_rm = [
    {'remove': {'alias': 'my_alias', 'index': 'index-2016.03.03'}},
    {'remove': {'alias': 'my_alias', 'index': 'index-2016.03.04'}},
]
alias_success = {"acknowledged": True}
allocation_in = {
    named_index: {
        'settings': {'index': {'routing': {'allocation': {'require': {'foo': 'bar'}}}}}
    }
}
allocation_out = {
    named_index: {
        'settings': {'index': {'routing': {'allocation': {'require': {'not': 'foo'}}}}}
    }
}
indices_space = {
    'indices': {
        'index1': {'index': {'primary_size_in_bytes': 1083741824}},
        'index2': {'index': {'primary_size_in_bytes': 1083741824}},
    }
}
snap_name = 'snap_name'
repo_name = 'repo_name'
test_repo = {
    repo_name: {
        'type': 'fs',
        'settings': {'compress': 'true', 'location': '/tmp/repos/repo_name'},
    }
}
test_repos = {
    'TESTING': {
        'type': 'fs',
        'settings': {'compress': 'true', 'location': '/tmp/repos/TESTING'},
    },
    repo_name: {
        'type': 'fs',
        'settings': {'compress': 'true', 'location': '/rmp/repos/repo_name'},
    },
}
snap_running = {'snapshots': ['running']}
nosnap_running = {'snapshots': []}
snapshot = {
    'snapshots': [
        {
            'duration_in_millis': 60000,
            'start_time': '2015-02-01T00:00:00.000Z',
            'shards': {'successful': 4, 'failed': 0, 'total': 4},
            'end_time_in_millis': 0,
            'state': 'SUCCESS',
            'snapshot': snap_name,
            'end_time': '2015-02-01T00:00:01.000Z',
            'indices': named_indices,
            'failures': [],
            'start_time_in_millis': 1422748800,
        }
    ]
}
oneinprogress = {
    'snapshots': [
        {
            'duration_in_millis': 60000,
            'start_time': '2015-03-01T00:00:02.000Z',
            'shards': {'successful': 4, 'failed': 0, 'total': 4},
            'end_time_in_millis': 0,
            'state': 'IN_PROGRESS',
            'snapshot': snap_name,
            'end_time': '2015-03-01T00:00:03.000Z',
            'indices': named_indices,
            'failures': [],
            'start_time_in_millis': 1425168002,
        }
    ]
}
partial = {
    'snapshots': [
        {
            'duration_in_millis': 60000,
            'start_time': '2015-02-01T00:00:00.000Z',
            'shards': {'successful': 4, 'failed': 0, 'total': 4},
            'end_time_in_millis': 0,
            'state': 'PARTIAL',
            'snapshot': snap_name,
            'end_time': '2015-02-01T00:00:01.000Z',
            'indices': named_indices,
            'failures': [],
            'start_time_in_millis': 1422748800,
        }
    ]
}
failed = {
    'snapshots': [
        {
            'duration_in_millis': 60000,
            'start_time': '2015-02-01T00:00:00.000Z',
            'shards': {'successful': 4, 'failed': 0, 'total': 4},
            'end_time_in_millis': 0,
            'state': 'FAILED',
            'snapshot': snap_name,
            'end_time': '2015-02-01T00:00:01.000Z',
            'indices': named_indices,
            'failures': [],
            'start_time_in_millis': 1422748800,
        }
    ]
}
othersnap = {
    'snapshots': [
        {
            'duration_in_millis': 60000,
            'start_time': '2015-02-01T00:00:00.000Z',
            'shards': {'successful': 4, 'failed': 0, 'total': 4},
            'end_time_in_millis': 0,
            'state': 'SOMETHINGELSE',
            'snapshot': snap_name,
            'end_time': '2015-02-01T00:00:01.000Z',
            'indices': named_indices,
            'failures': [],
            'start_time_in_millis': 1422748800,
        }
    ]
}
snapshots = {
    'snapshots': [
        {
            'duration_in_millis': 60000,
            'start_time': '2015-02-01T00:00:00.000Z',
            'shards': {'successful': 4, 'failed': 0, 'total': 4},
            'end_time_in_millis': 0,
            'state': 'SUCCESS',
            'snapshot': snap_name,
            'end_time': '2015-02-01T00:00:01.000Z',
            'indices': named_indices,
            'failures': [],
            'start_time_in_millis': 1422748800,
        },
        {
            'duration_in_millis': 60000,
            'start_time': '2015-03-01T00:00:02.000Z',
            'shards': {'successful': 4, 'failed': 0, 'total': 4},
            'end_time_in_millis': 0,
            'state': 'SUCCESS',
            'snapshot': 'snapshot-2015.03.01',
            'end_time': '2015-03-01T00:00:03.000Z',
            'indices': named_indices,
            'failures': [],
            'start_time_in_millis': 1425168002,
        },
    ]
}
inprogress = {
    'snapshots': [
        {
            'duration_in_millis': 60000,
            'start_time': '2015-02-01T00:00:00.000Z',
            'shards': {'successful': 4, 'failed': 0, 'total': 4},
            'end_time_in_millis': 0,
            'state': 'SUCCESS',
            'snapshot': snap_name,
            'end_time': '2015-02-01T00:00:01.000Z',
            'indices': named_indices,
            'failures': [],
            'start_time_in_millis': 1422748800,
        },
        {
            'duration_in_millis': 60000,
            'start_time': '2015-03-01T00:00:02.000Z',
            'shards': {'successful': 4, 'failed': 0, 'total': 4},
            'end_time_in_millis': 0,
            'state': 'IN_PROGRESS',
            'snapshot': 'snapshot-2015.03.01',
            'end_time': '2015-03-01T00:00:03.000Z',
            'indices': named_indices,
            'failures': [],
            'start_time_in_millis': 1425168002,
        },
    ]
}
highly_unlikely = {
    'snapshots': [
        {
            'duration_in_millis': 60000,
            'start_time': '2015-02-01T00:00:00.000Z',
            'shards': {'successful': 4, 'failed': 0, 'total': 4},
            'end_time_in_millis': 0,
            'state': 'IN_PROGRESS',
            'snapshot': snap_name,
            'end_time': '2015-02-01T00:00:01.000Z',
            'indices': named_indices,
            'failures': [],
            'start_time_in_millis': 1422748800,
        },
        {
            'duration_in_millis': 60000,
            'start_time': '2015-03-01T00:00:02.000Z',
            'shards': {'successful': 4, 'failed': 0, 'total': 4},
            'end_time_in_millis': 0,
            'state': 'IN_PROGRESS',
            'snapshot': 'snapshot-2015.03.01',
            'end_time': '2015-03-01T00:00:03.000Z',
            'indices': named_indices,
            'failures': [],
            'start_time_in_millis': 1425168002,
        },
    ]
}
snap_body_all = {
    "ignore_unavailable": False,
    "include_global_state": True,
    "partial": False,
    "indices": "_all",
}
snap_body = {
    "ignore_unavailable": False,
    "include_global_state": True,
    "partial": False,
    "indices": "index-2015.01.01,index-2015.02.01",
}
verified_nodes = {'nodes': {'nodeid1': {'name': 'node1'}, 'nodeid2': {'name': 'node2'}}}
synced_pass = {
    "_shards": {"total": 1, "successful": 1, "failed": 0},
    "index_name": {
        "total": 1,
        "successful": 1,
        "failed": 0,
        "failures": [],
    },
}
synced_fail = {
    "_shards": {"total": 1, "successful": 0, "failed": 1},
    "index_name": {
        "total": 1,
        "successful": 0,
        "failed": 1,
        "failures": [
            {
                "shard": 0,
                "reason": "pending operations",
                "routing": {
                    "state": "STARTED",
                    "primary": True,
                    "node": "nodeid1",
                    "relocating_node": None,
                    "shard": 0,
                    "index": "index_name",
                },
            },
        ],
    },
}
sync_conflict = ConflictError(
    409,
    '{"_shards":{"total":1,"successful":0,"failed":1},"index_name":{"total":1,"successful":0,"failed":1,"failures":[{"shard":0,"reason":"pending operations","routing":{"state":"STARTED","primary":true,"node":"nodeid1","relocating_node":null,"shard":0,"index":"index_name"}}]}})',
    synced_fail,
)
synced_fails = {
    "_shards": {"total": 2, "successful": 1, "failed": 1},
    "index1": {
        "total": 1,
        "successful": 0,
        "failed": 1,
        "failures": [
            {
                "shard": 0,
                "reason": "pending operations",
                "routing": {
                    "state": "STARTED",
                    "primary": True,
                    "node": "nodeid1",
                    "relocating_node": None,
                    "shard": 0,
                    "index": "index_name",
                },
            },
        ],
    },
    "index2": {"total": 1, "successful": 1, "failed": 0, "failures": []},
}

state_one = [{'index': named_index, 'status': 'open'}]

settings_one = {
    named_index: {
        'aliases': ['my_alias'],
        'mappings': {},
        'settings': {
            'index': {
                'number_of_replicas': '1',
                'uuid': 'random_uuid_string_here',
                'number_of_shards': '2',
                'creation_date': '1456963200172',
                'routing': {'allocation': {'include': {'tag': 'foo'}}},
                'version': {'created': '2020099'},
                'refresh_interval': '5s',
            }
        },
    }
}

settings_1_get_aliases = {named_index: {"aliases": {'my_alias': {}}}}

state_two = [
    {'index': 'index-2016.03.03', 'status': 'open'},
    {'index': 'index-2016.03.04', 'status': 'open'},
]

settings_two = {
    'index-2016.03.03': {
        'aliases': ['my_alias'],
        'mappings': {},
        'settings': {
            'index': {
                'number_of_replicas': '1',
                'uuid': 'random_uuid_string_here',
                'number_of_shards': '5',
                'creation_date': '1456963200172',
                'routing': {'allocation': {'include': {'tag': 'foo'}}},
                'version': {'created': '2020099'},
                'refresh_interval': '5s',
            }
        },
    },
    'index-2016.03.04': {
        'aliases': ['my_alias'],
        'mappings': {},
        'settings': {
            'index': {
                'number_of_replicas': '1',
                'uuid': 'another_random_uuid_string',
                'number_of_shards': '5',
                'creation_date': '1457049600812',
                'routing': {'allocation': {'include': {'tag': 'bar'}}},
                'version': {'created': '2020099'},
                'refresh_interval': '5s',
            }
        },
    },
}

settings_2_get_aliases = {
    "index-2016.03.03": {"aliases": {'my_alias': {}}},
    "index-2016.03.04": {"aliases": {'my_alias': {}}},
}

state_2_closed = [
    {'index': 'index-2016.03.03', 'status': 'close'},
    {'index': 'index-2016.03.04', 'status': 'open'},
]

settings_2_closed = {
    'index-2016.03.03': {
        'aliases': ['my_alias'],
        'mappings': {},
        'settings': {
            'index': {
                'number_of_replicas': '1',
                'uuid': 'random_uuid_string_here',
                'number_of_shards': '5',
                'creation_date': '1456963200172',
                'routing': {'allocation': {'include': {'tag': 'foo'}}},
                'version': {'created': '2020099'},
                'refresh_interval': '5s',
            }
        },
    },
    'index-2016.03.04': {
        'aliases': ['my_alias'],
        'mappings': {},
        'settings': {
            'index': {
                'number_of_replicas': '1',
                'uuid': 'another_random_uuid_string',
                'number_of_shards': '5',
                'creation_date': '1457049600812',
                'routing': {'allocation': {'include': {'tag': 'bar'}}},
                'version': {'created': '2020099'},
                'refresh_interval': '5s',
            }
        },
    },
}

state_four = [
    {'index': 'a-2016.03.03', 'status': 'open'},
    {'index': 'b-2016.03.04', 'status': 'open'},
    {'index': 'c-2016.03.05', 'status': 'close'},
    {'index': 'd-2016.03.06', 'status': 'open'},
]

settings_four = {
    'a-2016.03.03': {
        'aliases': ['my_alias'],
        'mappings': {},
        'settings': {
            'index': {
                'number_of_replicas': '1',
                'uuid': 'random_uuid_string_here',
                'number_of_shards': '5',
                'creation_date': '1456963200172',
                'routing': {'allocation': {'include': {'tag': 'foo'}}},
                'version': {'created': '2020099'},
                'refresh_interval': '5s',
            }
        },
    },
    'b-2016.03.04': {
        'aliases': ['my_alias'],
        'mappings': {},
        'settings': {
            'index': {
                'number_of_replicas': '1',
                'uuid': 'another_random_uuid_string',
                'number_of_shards': '5',
                'creation_date': '1457049600812',
                'routing': {'allocation': {'include': {'tag': 'bar'}}},
                'version': {'created': '2020099'},
                'refresh_interval': '5s',
            }
        },
    },
    'c-2016.03.05': {
        'aliases': ['my_alias'],
        'mappings': {},
        'settings': {
            'index': {
                'number_of_replicas': '1',
                'uuid': 'random_uuid_string_here',
                'number_of_shards': '5',
                'creation_date': '1457136000933',
                'routing': {'allocation': {'include': {'tag': 'foo'}}},
                'version': {'created': '2020099'},
                'refresh_interval': '5s',
            }
        },
    },
    'd-2016.03.06': {
        'aliases': ['my_alias'],
        'mappings': {},
        'settings': {
            'index': {
                'number_of_replicas': '1',
                'uuid': 'another_random_uuid_string',
                'number_of_shards': '5',
                'creation_date': '1457222400527',
                'routing': {'allocation': {'include': {'tag': 'bar'}}},
                'version': {'created': '2020099'},
                'refresh_interval': '5s',
            }
        },
    },
}

state_named = [
    {'index': 'index-2015.01.01', 'status': 'open'},
    {'index': 'index-2015.02.01', 'status': 'open'},
]

settings_named = {
    'index-2015.01.01': {
        'aliases': ['my_alias'],
        'mappings': {},
        'settings': {
            'index': {
                'number_of_replicas': '1',
                'uuid': 'random_uuid_string_here',
                'number_of_shards': '5',
                'creation_date': '1456963200172',
                'routing': {'allocation': {'include': {'tag': 'foo'}}},
                'version': {'created': '2020099'},
                'refresh_interval': '5s',
            }
        },
    },
    'index-2015.02.01': {
        'aliases': ['my_alias'],
        'mappings': {},
        'settings': {
            'index': {
                'number_of_replicas': '1',
                'uuid': 'another_random_uuid_string',
                'number_of_shards': '5',
                'creation_date': '1457049600812',
                'routing': {'allocation': {'include': {'tag': 'bar'}}},
                'version': {'created': '2020099'},
                'refresh_interval': '5s',
            }
        },
    },
}

stats_one = {
    'indices': {
        named_index: {
            'total': {
                'docs': {'count': 6374962, 'deleted': 0},
                'store': {'size_in_bytes': 1115219663, 'throttle_time_in_millis': 0},
            },
            'primaries': {
                'docs': {'count': 3187481, 'deleted': 0},
                'store': {'size_in_bytes': 557951789, 'throttle_time_in_millis': 0},
            },
        }
    }
}

stats_two = {
    'indices': {
        'index-2016.03.03': {
            'total': {
                'docs': {'count': 6374962, 'deleted': 0},
                'store': {'size_in_bytes': 1115219663, 'throttle_time_in_millis': 0},
            },
            'primaries': {
                'docs': {'count': 3187481, 'deleted': 0},
                'store': {'size_in_bytes': 557951789, 'throttle_time_in_millis': 0},
            },
        },
        'index-2016.03.04': {
            'total': {
                'docs': {'count': 6377544, 'deleted': 0},
                'store': {'size_in_bytes': 1120891046, 'throttle_time_in_millis': 0},
            },
            'primaries': {
                'docs': {'count': 3188772, 'deleted': 0},
                'store': {'size_in_bytes': 560677114, 'throttle_time_in_millis': 0},
            },
        },
    }
}

stats_four = {
    'indices': {
        'a-2016.03.03': {
            'total': {
                'docs': {'count': 6374962, 'deleted': 0},
                'store': {'size_in_bytes': 1115219663, 'throttle_time_in_millis': 0},
            },
            'primaries': {
                'docs': {'count': 3187481, 'deleted': 0},
                'store': {'size_in_bytes': 557951789, 'throttle_time_in_millis': 0},
            },
        },
        'b-2016.03.04': {
            'total': {
                'docs': {'count': 6377544, 'deleted': 0},
                'store': {'size_in_bytes': 1120891046, 'throttle_time_in_millis': 0},
            },
            'primaries': {
                'docs': {'count': 3188772, 'deleted': 0},
                'store': {'size_in_bytes': 560677114, 'throttle_time_in_millis': 0},
            },
        },
        # CLOSED, ergo, not present
        # 'c-2016.03.05': {
        #     'total': {
        #         'docs': {'count': 6266434, 'deleted': 0},
        #         'store': {'size_in_bytes': 1120882166, 'throttle_time_in_millis': 0}
        #     },
        #     'primaries': {
        #         'docs': {'count': 3133217, 'deleted': 0},
        #         'store': {'size_in_bytes': 560441083, 'throttle_time_in_millis': 0}
        #     }
        # },
        'd-2016.03.06': {
            'total': {
                'docs': {'count': 6266436, 'deleted': 0},
                'store': {'size_in_bytes': 1120882168, 'throttle_time_in_millis': 0},
            },
            'primaries': {
                'docs': {'count': 3133218, 'deleted': 0},
                'store': {'size_in_bytes': 560441084, 'throttle_time_in_millis': 0},
            },
        },
    }
}

fieldstats_one = {
    'indices': {
        named_index: {
            'fields': {
                'timestamp': {
                    'density': 100,
                    'min_value_as_string': '2016-03-03T00:00:06.189Z',
                    'max_value': 1457049599152,
                    'max_doc': 415651,
                    'min_value': 1456963206189,
                    'doc_count': 415651,
                    'max_value_as_string': '2016-03-03T23:59:59.152Z',
                    'sum_total_term_freq': -1,
                    'sum_doc_freq': 1662604,
                }
            }
        }
    }
}

fieldstats_two = {
    'indices': {
        'index-2016.03.03': {
            'fields': {
                'timestamp': {
                    'density': 100,
                    'min_value_as_string': '2016-03-03T00:00:06.189Z',
                    'max_value': 1457049599152,
                    'max_doc': 415651,
                    'min_value': 1456963206189,
                    'doc_count': 415651,
                    'max_value_as_string': '2016-03-03T23:59:59.152Z',
                    'sum_total_term_freq': -1,
                    'sum_doc_freq': 1662604,
                }
            }
        },
        'index-2016.03.04': {
            'fields': {
                'timestamp': {
                    'density': 100,
                    'min_value_as_string': '2016-03-04T00:00:00.812Z',
                    'max_value': 1457135999223,
                    'max_doc': 426762,
                    'min_value': 1457049600812,
                    'doc_count': 426762,
                    'max_value_as_string': '2016-03-04T23:59:59.223Z',
                    'sum_total_term_freq': -1,
                    'sum_doc_freq': 1673715,
                }
            }
        },
    }
}

fieldstats_four = {
    'indices': {
        'a-2016.03.03': {
            'fields': {
                'timestamp': {
                    'density': 100,
                    'min_value_as_string': '2016-03-03T00:00:06.189Z',
                    'max_value': 1457049599152,
                    'max_doc': 415651,
                    'min_value': 1456963206189,
                    'doc_count': 415651,
                    'max_value_as_string': '2016-03-03T23:59:59.152Z',
                    'sum_total_term_freq': -1,
                    'sum_doc_freq': 1662604,
                }
            }
        },
        'b-2016.03.04': {
            'fields': {
                'timestamp': {
                    'density': 100,
                    'min_value_as_string': '2016-03-04T00:00:00.812Z',
                    'max_value': 1457135999223,
                    'max_doc': 426762,
                    'min_value': 1457049600812,
                    'doc_count': 426762,
                    'max_value_as_string': '2016-03-04T23:59:59.223Z',
                    'sum_total_term_freq': -1,
                    'sum_doc_freq': 1673715,
                }
            }
        },
        'd-2016.03.06': {
            'fields': {
                'timestamp': {
                    'density': 100,
                    'min_value_as_string': '2016-03-04T00:00:00.812Z',
                    'max_value': 1457308799223,
                    'max_doc': 426762,
                    'min_value': 1457222400567,
                    'doc_count': 426762,
                    'max_value_as_string': '2016-03-04T23:59:59.223Z',
                    'sum_total_term_freq': -1,
                    'sum_doc_freq': 1673715,
                }
            }
        },
    }
}

fieldstats_query = {
    'aggregations': {
        'min': {
            'value_as_string': '2016-03-03T00:00:06.189Z',
            'value': 1456963206189,
        },
        'max': {
            'value': 1457049599152,
            'value_as_string': '2016-03-03T23:59:59.152Z',
        },
    }
}

shards = {
    'indices': {
        named_index: {
            'shards': {
                '0': [{'num_search_segments': 15}, {'num_search_segments': 21}],
                '1': [{'num_search_segments': 19}, {'num_search_segments': 16}],
            }
        }
    }
}
fm_shards = {
    'indices': {
        named_index: {
            'shards': {
                '0': [{'num_search_segments': 1}, {'num_search_segments': 1}],
                '1': [{'num_search_segments': 1}, {'num_search_segments': 1}],
            }
        }
    }
}

loginfo = {"loglevel": "INFO", "logfile": None, "logformat": "default"}
default_format = '%(asctime)s %(levelname)-9s %(message)s'
debug_format = (
    '%(asctime)s %(levelname)-9s %(name)22s %(funcName)22s:%(lineno)-4d %(message)s'
)

yamlconfig = '''
---
# Remember, leave a key empty to use the default value.  None will be a string,
# not a Python "NoneType"
client:
  hosts: http://127.0.0.1:9200
  certificate:
  client_cert:
  client_key:
  http_auth:
  timeout: 30
  master_only: False

options:
  dry_run: False
  loglevel: DEBUG
  logfile:
  logformat: default
  quiet: False
'''
pattern_ft = '''
---
actions:
  1:
    description: open all matching indices
    action: open
    options:
      continue_if_exception: False
      disable_action: False
    filters:
      - filtertype: pattern
        kind: prefix
        value: a
        exclude: False
'''
age_ft = '''
---
actions:
  1:
    description: open all matching indices
    action: open
    options:
      continue_if_exception: False
      disable_action: False
    filters:
      - filtertype: age
        source: name
        direction: older
        timestring: '%Y.%m.%d'
        unit: seconds
        unit_count: 0
        epoch: 1456963201
'''
space_ft = '''
---
actions:
  1:
    description: open all matching indices
    action: open
    options:
      continue_if_exception: False
      disable_action: False
    filters:
      - filtertype: space
        disk_space: 2.1
        source: name
        use_age: True
        timestring: '%Y.%m.%d'
'''
forcemerge_ft = '''
---
actions:
  1:
    description: open all matching indices
    action: open
    options:
      continue_if_exception: False
      disable_action: False
    filters:
      - filtertype: forcemerged
        max_num_segments: 2
'''
allocated_ft = '''
---
actions:
  1:
    description: open all matching indices
    action: open
    options:
      continue_if_exception: False
      disable_action: False
    filters:
      - filtertype: allocated
        key: tag
        value: foo
        allocation_type: include
'''
kibana_ft = '''
---
actions:
  1:
    description: open all matching indices
    action: open
    options:
      continue_if_exception: False
      disable_action: False
    filters:
      - filtertype: kibana
'''
opened_ft = '''
---
actions:
  1:
    description: open all matching indices
    action: open
    options:
      continue_if_exception: False
      disable_action: False
    filters:
      - filtertype: opened
'''
closed_ft = '''
---
actions:
  1:
    description: open all matching indices
    action: open
    options:
      continue_if_exception: False
      disable_action: False
    filters:
      - filtertype: closed
'''
none_ft = '''
---
actions:
  1:
    description: open all matching indices
    action: open
    options:
      continue_if_exception: False
      disable_action: False
    filters:
      - filtertype: none
'''
invalid_ft = '''
---
actions:
  1:
    description: open all matching indices
    action: open
    options:
      continue_if_exception: False
      disable_action: False
    filters:
      - filtertype: sir_not_appearing_in_this_film
'''
snap_age_ft = '''
---
actions:
  1:
    description: test
    action: delete_snapshots
    options:
      continue_if_exception: False
      disable_action: False
    filters:
      - filtertype: age
        direction: older
        unit: days
        unit_count: 1
'''
snap_pattern_ft = '''
---
actions:
  1:
    description: test
    action: delete_snapshots
    options:
      continue_if_exception: False
      disable_action: False
    filters:
      - filtertype: pattern
        kind: prefix
        value: sna
'''
snap_none_ft = '''
---
actions:
  1:
    description: test
    action: delete_snapshots
    options:
      continue_if_exception: False
      disable_action: False
    filters:
      - filtertype: none
'''
size_ft = '''
---
actions:
  1:
    description: open all matching indices
    action: open
    options:
      continue_if_exception: False
      disable_action: False
    filters:
      - filtertype: size
        size_threshold: 1.04
        size_behavior: total
        threshold_behavior: less_than
'''

generic_task = {'task': 'I0ekFjMhSPCQz7FUs1zJOg:54510686'}
incomplete_task = {
    'completed': False,
    'task': {
        'node': 'I0ekFjMhSPCQz7FUs1zJOg',
        'status': {
            'retries': {'bulk': 0, 'search': 0},
            'updated': 0,
            'batches': 3647,
            'throttled_until_millis': 0,
            'throttled_millis': 0,
            'noops': 0,
            'created': 3646581,
            'deleted': 0,
            'requests_per_second': -1.0,
            'version_conflicts': 0,
            'total': 3646581,
        },
        'description': 'UNIT TEST',
        'running_time_in_nanos': 1637039537721,
        'cancellable': True,
        'action': 'indices:data/write/reindex',
        'type': 'transport',
        'id': 54510686,
        'start_time_in_millis': 1489695981997,
    },
    'response': {
        'retries': {'bulk': 0, 'search': 0},
        'updated': 0,
        'batches': 3647,
        'throttled_until_millis': 0,
        'throttled_millis': 0,
        'noops': 0,
        'created': 3646581,
        'deleted': 0,
        'took': 1636917,
        'requests_per_second': -1.0,
        'timed_out': False,
        'failures': [],
        'version_conflicts': 0,
        'total': 3646581,
    },
}
completed_task = {
    'completed': True,
    'task': {
        'node': 'I0ekFjMhSPCQz7FUs1zJOg',
        'status': {
            'retries': {'bulk': 0, 'search': 0},
            'updated': 0,
            'batches': 3647,
            'throttled_until_millis': 0,
            'throttled_millis': 0,
            'noops': 0,
            'created': 3646581,
            'deleted': 0,
            'requests_per_second': -1.0,
            'version_conflicts': 0,
            'total': 3646581,
        },
        'description': 'UNIT TEST',
        'running_time_in_nanos': 1637039537721,
        'cancellable': True,
        'action': 'indices:data/write/reindex',
        'type': 'transport',
        'id': 54510686,
        'start_time_in_millis': 1489695981997,
    },
    'response': {
        'retries': {'bulk': 0, 'search': 0},
        'updated': 0,
        'batches': 3647,
        'throttled_until_millis': 0,
        'throttled_millis': 0,
        'noops': 0,
        'created': 3646581,
        'deleted': 0,
        'took': 1636917,
        'requests_per_second': -1.0,
        'timed_out': False,
        'failures': [],
        'version_conflicts': 0,
        'total': 3646581,
    },
}
completed_task_zero_total = {
    'completed': True,
    'task': {
        'node': 'I0ekFjMhSPCQz7FUs1zJOg',
        'status': {
            'retries': {'bulk': 0, 'search': 0},
            'updated': 0,
            'batches': 0,
            'throttled_until_millis': 0,
            'throttled_millis': 0,
            'noops': 0,
            'created': 0,
            'deleted': 0,
            'requests_per_second': -1.0,
            'version_conflicts': 0,
            'total': 0,
        },
        'description': 'UNIT TEST',
        'running_time_in_nanos': 1637039537721,
        'cancellable': True,
        'action': 'indices:data/write/reindex',
        'type': 'transport',
        'id': 54510686,
        'start_time_in_millis': 1489695981997,
    },
    'response': {
        'retries': {'bulk': 0, 'search': 0},
        'updated': 0,
        'batches': 0,
        'throttled_until_millis': 0,
        'throttled_millis': 0,
        'noops': 0,
        'created': 0,
        'deleted': 0,
        'took': 1636917,
        'requests_per_second': -1.0,
        'timed_out': False,
        'failures': [],
        'version_conflicts': 0,
        'total': 0,
    },
}
recovery_output = {
    'index-2015.01.01': {'shards': [{'stage': 'DONE'}]},
    'index-2015.02.01': {'shards': [{'stage': 'DONE'}]},
}
unrecovered_output = {
    'index-2015.01.01': {'shards': [{'stage': 'INDEX'}]},
    'index-2015.02.01': {'shards': [{'stage': 'INDEX'}]},
}
cluster_health = {
    "cluster_name": "unit_test",
    "status": "green",
    "timed_out": False,
    "number_of_nodes": 7,
    "number_of_data_nodes": 3,
    "active_primary_shards": 235,
    "active_shards": 471,
    "relocating_shards": 0,
    "initializing_shards": 0,
    "unassigned_shards": 0,
    "delayed_unassigned_shards": 0,
    "number_of_pending_tasks": 0,
    "task_max_waiting_in_queue_millis": 0,
    "active_shards_percent_as_number": 100,
}
reindex_basic = {'source': {'index': named_index}, 'dest': {'index': 'other_index'}}
reindex_replace = {
    'source': {'index': 'REINDEX_SELECTION'},
    'dest': {'index': 'other_index'},
}
reindex_migration = {'source': {'index': named_index}, 'dest': {'index': 'MIGRATION'}}
index_list_966 = ['indexv0.2_2017-02-12_536a9247f9fa4fc7a7942ad46ea14e0d']
recovery_966 = {
    'indexv0.2_2017-02-12_536a9247f9fa4fc7a7942ad46ea14e0d': {
        'shards': [
            {
                'total_time': '10.1m',
                'index': {
                    'files': {
                        'reused': 0,
                        'total': 15,
                        'percent': '100.0%',
                        'recovered': 15,
                    },
                    'total_time': '10.1m',
                    'target_throttle_time': '-1',
                    'total_time_in_millis': 606577,
                    'source_throttle_time_in_millis': 0,
                    'source_throttle_time': '-1',
                    'target_throttle_time_in_millis': 0,
                    'size': {
                        'recovered_in_bytes': 3171596177,
                        'reused': '0b',
                        'total_in_bytes': 3171596177,
                        'percent': '100.0%',
                        'reused_in_bytes': 0,
                        'total': '2.9gb',
                        'recovered': '2.9gb',
                    },
                },
                'verify_index': {
                    'total_time': '0s',
                    'total_time_in_millis': 0,
                    'check_index_time_in_millis': 0,
                    'check_index_time': '0s',
                },
                'target': {
                    'ip': 'x.x.x.7',
                    'host': 'x.x.x.7',
                    'transport_address': 'x.x.x.7:9300',
                    'id': 'K4xQPaOFSWSPLwhb0P47aQ',
                    'name': 'staging-es5-forcem',
                },
                'source': {
                    'index': 'indexv0.2_2017-02-12_536a9247f9fa4fc7a7942ad46ea14e0d',
                    'version': '5.1.1',
                    'snapshot': 'force-merge',
                    'repository': 'force-merge',
                },
                'translog': {
                    'total_time': '45ms',
                    'percent': '100.0%',
                    'total_time_in_millis': 45,
                    'total_on_start': 0,
                    'total': 0,
                    'recovered': 0,
                },
                'start_time': '2017-05-16T11:54:48.183Z',
                'primary': True,
                'total_time_in_millis': 606631,
                'stop_time_in_millis': 1494936294815,
                'stop_time': '2017-05-16T12:04:54.815Z',
                'stage': 'DONE',
                'type': 'SNAPSHOT',
                'id': 1,
                'start_time_in_millis': 1494935688183,
            },
            {
                'total_time': '10m',
                'index': {
                    'files': {
                        'reused': 0,
                        'total': 15,
                        'percent': '100.0%',
                        'recovered': 15,
                    },
                    'total_time': '10m',
                    'target_throttle_time': '-1',
                    'total_time_in_millis': 602302,
                    'source_throttle_time_in_millis': 0,
                    'source_throttle_time': '-1',
                    'target_throttle_time_in_millis': 0,
                    'size': {
                        'recovered_in_bytes': 3162299781,
                        'reused': '0b',
                        'total_in_bytes': 3162299781,
                        'percent': '100.0%',
                        'reused_in_bytes': 0,
                        'total': '2.9gb',
                        'recovered': '2.9gb',
                    },
                },
                'verify_index': {
                    'total_time': '0s',
                    'total_time_in_millis': 0,
                    'check_index_time_in_millis': 0,
                    'check_index_time': '0s',
                },
                'target': {
                    'ip': 'x.x.x.7',
                    'host': 'x.x.x.7',
                    'transport_address': 'x.x.x.7:9300',
                    'id': 'K4xQPaOFSWSPLwhb0P47aQ',
                    'name': 'staging-es5-forcem',
                },
                'source': {
                    'index': 'indexv0.2_2017-02-12_536a9247f9fa4fc7a7942ad46ea14e0d',
                    'version': '5.1.1',
                    'snapshot': 'force-merge',
                    'repository': 'force-merge',
                },
                'translog': {
                    'total_time': '389ms',
                    'percent': '100.0%',
                    'total_time_in_millis': 389,
                    'total_on_start': 0,
                    'total': 0,
                    'recovered': 0,
                },
                'start_time': '2017-05-16T12:04:51.606Z',
                'primary': True,
                'total_time_in_millis': 602698,
                'stop_time_in_millis': 1494936894305,
                'stop_time': '2017-05-16T12:14:54.305Z',
                'stage': 'DONE',
                'type': 'SNAPSHOT',
                'id': 5,
                'start_time_in_millis': 1494936291606,
            },
            {
                'total_time': '10.1m',
                'index': {
                    'files': {
                        'reused': 0,
                        'total': 15,
                        'percent': '100.0%',
                        'recovered': 15,
                    },
                    'total_time': '10.1m',
                    'target_throttle_time': '-1',
                    'total_time_in_millis': 606692,
                    'source_throttle_time_in_millis': 0,
                    'source_throttle_time': '-1',
                    'target_throttle_time_in_millis': 0,
                    'size': {
                        'recovered_in_bytes': 3156050994,
                        'reused': '0b',
                        'total_in_bytes': 3156050994,
                        'percent': '100.0%',
                        'reused_in_bytes': 0,
                        'total': '2.9gb',
                        'recovered': '2.9gb',
                    },
                },
                'verify_index': {
                    'total_time': '0s',
                    'total_time_in_millis': 0,
                    'check_index_time_in_millis': 0,
                    'check_index_time': '0s',
                },
                'target': {
                    'ip': 'x.x.x.7',
                    'host': 'x.x.x.7',
                    'transport_address': 'x.x.x.7:9300',
                    'id': 'K4xQPaOFSWSPLwhb0P47aQ',
                    'name': 'staging-es5-forcem',
                },
                'source': {
                    'index': 'indexv0.2_2017-02-12_536a9247f9fa4fc7a7942ad46ea14e0d',
                    'version': '5.1.1',
                    'snapshot': 'force-merge',
                    'repository': 'force-merge',
                },
                'translog': {
                    'total_time': '38ms',
                    'percent': '100.0%',
                    'total_time_in_millis': 38,
                    'total_on_start': 0,
                    'total': 0,
                    'recovered': 0,
                },
                'start_time': '2017-05-16T11:54:48.166Z',
                'primary': True,
                'total_time_in_millis': 606737,
                'stop_time_in_millis': 1494936294904,
                'stop_time': '2017-05-16T12:04:54.904Z',
                'stage': 'DONE',
                'type': 'SNAPSHOT',
                'id': 3,
                'start_time_in_millis': 1494935688166,
            },
            {
                'total_time': '10m',
                'index': {
                    'files': {
                        'reused': 0,
                        'total': 15,
                        'percent': '100.0%',
                        'recovered': 15,
                    },
                    'total_time': '10m',
                    'target_throttle_time': '-1',
                    'total_time_in_millis': 602010,
                    'source_throttle_time_in_millis': 0,
                    'source_throttle_time': '-1',
                    'target_throttle_time_in_millis': 0,
                    'size': {
                        'recovered_in_bytes': 3153017440,
                        'reused': '0b',
                        'total_in_bytes': 3153017440,
                        'percent': '100.0%',
                        'reused_in_bytes': 0,
                        'total': '2.9gb',
                        'recovered': '2.9gb',
                    },
                },
                'verify_index': {
                    'total_time': '0s',
                    'total_time_in_millis': 0,
                    'check_index_time_in_millis': 0,
                    'check_index_time': '0s',
                },
                'target': {
                    'ip': 'x.x.x.7',
                    'host': 'x.x.x.7',
                    'transport_address': 'x.x.x.7:9300',
                    'id': 'K4xQPaOFSWSPLwhb0P47aQ',
                    'name': 'staging-es5-forcem',
                },
                'source': {
                    'index': 'indexv0.2_2017-02-12_536a9247f9fa4fc7a7942ad46ea14e0d',
                    'version': '5.1.1',
                    'snapshot': 'force-merge',
                    'repository': 'force-merge',
                },
                'translog': {
                    'total_time': '558ms',
                    'percent': '100.0%',
                    'total_time_in_millis': 558,
                    'total_on_start': 0,
                    'total': 0,
                    'recovered': 0,
                },
                'start_time': '2017-05-16T12:04:51.369Z',
                'primary': True,
                'total_time_in_millis': 602575,
                'stop_time_in_millis': 1494936893944,
                'stop_time': '2017-05-16T12:14:53.944Z',
                'stage': 'DONE',
                'type': 'SNAPSHOT',
                'id': 4,
                'start_time_in_millis': 1494936291369,
            },
            {
                'total_time': '10m',
                'index': {
                    'files': {
                        'reused': 0,
                        'total': 15,
                        'percent': '100.0%',
                        'recovered': 15,
                    },
                    'total_time': '10m',
                    'target_throttle_time': '-1',
                    'total_time_in_millis': 600492,
                    'source_throttle_time_in_millis': 0,
                    'source_throttle_time': '-1',
                    'target_throttle_time_in_millis': 0,
                    'size': {
                        'recovered_in_bytes': 3153347402,
                        'reused': '0b',
                        'total_in_bytes': 3153347402,
                        'percent': '100.0%',
                        'reused_in_bytes': 0,
                        'total': '2.9gb',
                        'recovered': '2.9gb',
                    },
                },
                'verify_index': {
                    'total_time': '0s',
                    'total_time_in_millis': 0,
                    'check_index_time_in_millis': 0,
                    'check_index_time': '0s',
                },
                'target': {
                    'ip': 'x.x.x.7',
                    'host': 'x.x.x.7',
                    'transport_address': 'x.x.x.7:9300',
                    'id': 'K4xQPaOFSWSPLwhb0P47aQ',
                    'name': 'staging-es5-forcem',
                },
                'source': {
                    'index': 'indexv0.2_2017-02-12_536a9247f9fa4fc7a7942ad46ea14e0d',
                    'version': '5.1.1',
                    'snapshot': 'force-merge',
                    'repository': 'force-merge',
                },
                'translog': {
                    'total_time': '445ms',
                    'percent': '100.0%',
                    'total_time_in_millis': 445,
                    'total_on_start': 0,
                    'total': 0,
                    'recovered': 0,
                },
                'start_time': '2017-05-16T12:04:54.817Z',
                'primary': True,
                'total_time_in_millis': 600946,
                'stop_time_in_millis': 1494936895764,
                'stop_time': '2017-05-16T12:14:55.764Z',
                'stage': 'DONE',
                'type': 'SNAPSHOT',
                'id': 6,
                'start_time_in_millis': 1494936294817,
            },
            {
                'total_time': '10m',
                'index': {
                    'files': {
                        'reused': 0,
                        'total': 15,
                        'percent': '100.0%',
                        'recovered': 15,
                    },
                    'total_time': '10m',
                    'target_throttle_time': '-1',
                    'total_time_in_millis': 603194,
                    'source_throttle_time_in_millis': 0,
                    'source_throttle_time': '-1',
                    'target_throttle_time_in_millis': 0,
                    'size': {
                        'recovered_in_bytes': 3148003580,
                        'reused': '0b',
                        'total_in_bytes': 3148003580,
                        'percent': '100.0%',
                        'reused_in_bytes': 0,
                        'total': '2.9gb',
                        'recovered': '2.9gb',
                    },
                },
                'verify_index': {
                    'total_time': '0s',
                    'total_time_in_millis': 0,
                    'check_index_time_in_millis': 0,
                    'check_index_time': '0s',
                },
                'target': {
                    'ip': 'x.x.x.7',
                    'host': 'x.x.x.7',
                    'transport_address': 'x.x.x.7:9300',
                    'id': 'K4xQPaOFSWSPLwhb0P47aQ',
                    'name': 'staging-es5-forcem',
                },
                'source': {
                    'index': 'indexv0.2_2017-02-12_536a9247f9fa4fc7a7942ad46ea14e0d',
                    'version': '5.1.1',
                    'snapshot': 'force-merge',
                    'repository': 'force-merge',
                },
                'translog': {
                    'total_time': '225ms',
                    'percent': '100.0%',
                    'total_time_in_millis': 225,
                    'total_on_start': 0,
                    'total': 0,
                    'recovered': 0,
                },
                'start_time': '2017-05-16T11:54:48.173Z',
                'primary': True,
                'total_time_in_millis': 603429,
                'stop_time_in_millis': 1494936291602,
                'stop_time': '2017-05-16T12:04:51.602Z',
                'stage': 'DONE',
                'type': 'SNAPSHOT',
                'id': 2,
                'start_time_in_millis': 1494935688173,
            },
            {
                'total_time': '10m',
                'index': {
                    'files': {
                        'reused': 0,
                        'total': 15,
                        'percent': '100.0%',
                        'recovered': 15,
                    },
                    'total_time': '10m',
                    'target_throttle_time': '-1',
                    'total_time_in_millis': 601453,
                    'source_throttle_time_in_millis': 0,
                    'source_throttle_time': '-1',
                    'target_throttle_time_in_millis': 0,
                    'size': {
                        'recovered_in_bytes': 3168132171,
                        'reused': '0b',
                        'total_in_bytes': 3168132171,
                        'percent': '100.0%',
                        'reused_in_bytes': 0,
                        'total': '2.9gb',
                        'recovered': '2.9gb',
                    },
                },
                'verify_index': {
                    'total_time': '0s',
                    'total_time_in_millis': 0,
                    'check_index_time_in_millis': 0,
                    'check_index_time': '0s',
                },
                'target': {
                    'ip': 'x.x.x.7',
                    'host': 'x.x.x.7',
                    'transport_address': 'x.x.x.7:9300',
                    'id': 'K4xQPaOFSWSPLwhb0P47aQ',
                    'name': 'staging-es5-forcem',
                },
                'source': {
                    'index': 'indexv0.2_2017-02-12_536a9247f9fa4fc7a7942ad46ea14e0d',
                    'version': '5.1.1',
                    'snapshot': 'force-merge',
                    'repository': 'force-merge',
                },
                'translog': {
                    'total_time': '43ms',
                    'percent': '100.0%',
                    'total_time_in_millis': 43,
                    'total_on_start': 0,
                    'total': 0,
                    'recovered': 0,
                },
                'start_time': '2017-05-16T12:04:54.905Z',
                'primary': True,
                'total_time_in_millis': 601503,
                'stop_time_in_millis': 1494936896408,
                'stop_time': '2017-05-16T12:14:56.408Z',
                'stage': 'DONE',
                'type': 'SNAPSHOT',
                'id': 7,
                'start_time_in_millis': 1494936294905,
            },
            {
                'total_time': '10m',
                'index': {
                    'files': {
                        'reused': 0,
                        'total': 15,
                        'percent': '100.0%',
                        'recovered': 15,
                    },
                    'total_time': '10m',
                    'target_throttle_time': '-1',
                    'total_time_in_millis': 602897,
                    'source_throttle_time_in_millis': 0,
                    'source_throttle_time': '-1',
                    'target_throttle_time_in_millis': 0,
                    'size': {
                        'recovered_in_bytes': 3153750393,
                        'reused': '0b',
                        'total_in_bytes': 3153750393,
                        'percent': '100.0%',
                        'reused_in_bytes': 0,
                        'total': '2.9gb',
                        'recovered': '2.9gb',
                    },
                },
                'verify_index': {
                    'total_time': '0s',
                    'total_time_in_millis': 0,
                    'check_index_time_in_millis': 0,
                    'check_index_time': '0s',
                },
                'target': {
                    'ip': 'x.x.x.7',
                    'host': 'x.x.x.7',
                    'transport_address': 'x.x.x.7:9300',
                    'id': 'K4xQPaOFSWSPLwhb0P47aQ',
                    'name': 'staging-es5-forcem',
                },
                'source': {
                    'index': 'indexv0.2_2017-02-12_536a9247f9fa4fc7a7942ad46ea14e0d',
                    'version': '5.1.1',
                    'snapshot': 'force-merge',
                    'repository': 'force-merge',
                },
                'translog': {
                    'total_time': '271ms',
                    'percent': '100.0%',
                    'total_time_in_millis': 271,
                    'total_on_start': 0,
                    'total': 0,
                    'recovered': 0,
                },
                'start_time': '2017-05-16T11:54:48.191Z',
                'primary': True,
                'total_time_in_millis': 603174,
                'stop_time_in_millis': 1494936291366,
                'stop_time': '2017-05-16T12:04:51.366Z',
                'stage': 'DONE',
                'type': 'SNAPSHOT',
                'id': 0,
                'start_time_in_millis': 1494935688191,
            },
        ]
    }
}
no_snap_tasks = {
    'nodes': {
        'node1': {'tasks': {'task1': {'action': 'cluster:monitor/tasks/lists[n]'}}}
    }
}
snap_task = {
    'nodes': {
        'node1': {'tasks': {'task1': {'action': 'cluster:admin/snapshot/delete'}}}
    }
}
watermark_persistent = {
    'persistent': {
        'cluster': {
            'routing': {
                'allocation': {'disk': {'watermark': {'low': '11%', 'high': '60gb'}}}
            }
        }
    }
}
watermark_transient = {
    'transient': {
        'cluster': {
            'routing': {
                'allocation': {'disk': {'watermark': {'low': '9%', 'high': '50gb'}}}
            }
        }
    }
}
watermark_both = {
    'persistent': {
        'cluster': {
            'routing': {
                'allocation': {'disk': {'watermark': {'low': '11%', 'high': '60gb'}}}
            }
        }
    },
    'transient': {
        'cluster': {
            'routing': {
                'allocation': {'disk': {'watermark': {'low': '9%', 'high': '50gb'}}}
            }
        }
    },
}
empty_cluster_settings = {'persistent': {}, 'transient': {}}
data_only_node_role = ['data']
master_data_node_role = ['data', 'master']
